import tributary.streaming as ts


class TestAPI:
    def test_api(self):
        assert hasattr(ts.StreamingNode, '__add__')
        assert hasattr(ts.StreamingNode, '__add__')
        assert hasattr(ts.StreamingNode, '__radd__')
        assert hasattr(ts.StreamingNode, '__sub__')
        assert hasattr(ts.StreamingNode, '__rsub__')
        assert hasattr(ts.StreamingNode, '__mul__')
        assert hasattr(ts.StreamingNode, '__rmul__')
        assert hasattr(ts.StreamingNode, '__div__')
        assert hasattr(ts.StreamingNode, '__rdiv__')
        assert hasattr(ts.StreamingNode, '__truediv__')
        assert hasattr(ts.StreamingNode, '__rtruediv__')
        assert hasattr(ts.StreamingNode, '__pow__')
        assert hasattr(ts.StreamingNode, '__rpow__')
        assert hasattr(ts.StreamingNode, '__mod__')
        assert hasattr(ts.StreamingNode, '__rmod__')
        assert hasattr(ts.StreamingNode, 'sum')
        assert hasattr(ts.StreamingNode, 'average')
        assert hasattr(ts.StreamingNode, 'invert')
        assert hasattr(ts.StreamingNode, '__and__')
        assert hasattr(ts.StreamingNode, '__or__')
        assert hasattr(ts.StreamingNode, '__invert__')
        assert hasattr(ts.StreamingNode, '__bool__')
        assert hasattr(ts.StreamingNode, 'int')
        assert hasattr(ts.StreamingNode, 'float')
        assert hasattr(ts.StreamingNode, '__str__')
        assert hasattr(ts.StreamingNode, '__lt__')
        assert hasattr(ts.StreamingNode, '__le__')
        assert hasattr(ts.StreamingNode, '__gt__')
        assert hasattr(ts.StreamingNode, '__ge__')
        assert hasattr(ts.StreamingNode, '__eq__')
        assert hasattr(ts.StreamingNode, '__ne__')
        assert hasattr(ts.StreamingNode, '__neg__')
        assert hasattr(ts.StreamingNode, '__len__')
        assert hasattr(ts.StreamingNode, 'log')
        assert hasattr(ts.StreamingNode, 'sin')
        assert hasattr(ts.StreamingNode, 'cos')
        assert hasattr(ts.StreamingNode, 'tan')
        assert hasattr(ts.StreamingNode, 'asin')
        assert hasattr(ts.StreamingNode, 'acos')
        assert hasattr(ts.StreamingNode, 'atan')
        assert hasattr(ts.StreamingNode, 'abs')
        assert hasattr(ts.StreamingNode, 'sqrt')
        assert hasattr(ts.StreamingNode, 'exp')
        assert hasattr(ts.StreamingNode, 'erf')
        assert hasattr(ts.StreamingNode, 'round')
        assert hasattr(ts.StreamingNode, 'floor')
        assert hasattr(ts.StreamingNode, 'ceil')

        assert hasattr(ts.StreamingNode, 'delay')
        # assert hasattr(ts.StreamingNode, 'state')
        assert hasattr(ts.StreamingNode, 'apply')
        assert hasattr(ts.StreamingNode, 'window')
        assert hasattr(ts.StreamingNode, 'unroll')
        assert hasattr(ts.StreamingNode, 'unrollDataFrame')
        assert hasattr(ts.StreamingNode, 'merge')
        assert hasattr(ts.StreamingNode, 'listMerge')
        assert hasattr(ts.StreamingNode, 'dictMerge')
        assert hasattr(ts.StreamingNode, 'reduce')

        assert hasattr(ts.StreamingNode, 'graph')
        assert hasattr(ts.StreamingNode, 'pprint')
        assert hasattr(ts.StreamingNode, 'graphviz')
        assert hasattr(ts.StreamingNode, 'dagre')
        assert hasattr(ts.StreamingNode, 'print')

    def test_api_streaming_specific(self):
        '''apis specific to streaming nodes'''
        assert hasattr(ts.StreamingNode, 'perspective')
        assert hasattr(ts.StreamingNode, 'rollingCount')
        assert hasattr(ts.StreamingNode, 'rollingSum')
        assert hasattr(ts.StreamingNode, 'rollingCount')
        assert hasattr(ts.StreamingNode, 'rollingMin')
        assert hasattr(ts.StreamingNode, 'rollingMax')
        assert hasattr(ts.StreamingNode, 'rollingAverage')



import functools


def _either_type(f):
    '''Utility decorator to allow for either no-arg decorator or arg decorator

    Args:
        f (callable): Callable to decorate
    '''
    @functools.wraps(f)
    def new_dec(*args, **kwargs):
        if len(args) == 1 and len(kwargs) == 0 and callable(args[0]):
            # actual decorated function
            return f(args[0])
        else:
            # decorator arguments
            return lambda realf: f(realf, *args, **kwargs)
    return new_dec


def LazyToStreaming(lazy_node):
    from .streaming import StreamingNode, Foo
    from .lazy import LazyNode
    if isinstance(lazy_node, StreamingNode):
        return lazy_node
    if not isinstance(lazy_node, LazyNode):
        raise Exception("Malformed input:{}".format(lazy_node))

    return Foo(foo=lambda node=lazy_node: node())



from gevent import monkey
_PATCHED = False

if not _PATCHED:
    monkey.patch_all(thread=False, select=False)
    _PATCHED = True


from ._version import __version__  # noqa: F401

from .lazy import LazyGraph, LazyNode, node  # noqa: F401
from .functional import pipeline, stop  # noqa: F401
from .streaming import *  # noqa: F401, F403
from .utils import LazyToStreaming  # noqa: F401



# *****************************************************************************
#
# Copyright (c) 2019, the Perspective Authors.
#
# This file is part of the Perspective library, distributed under the terms of
# the Apache License 2.0.  The full license can be found in the LICENSE file.
#
import os
import time
import pytz
import numpy as np
import pandas as pd
from datetime import datetime
from dateutil import tz
from perspective.table import Table

LOCAL_DATETIMES = [
    datetime(2019, 1, 11, 0, 10, 20),
    datetime(2019, 1, 11, 11, 10, 20),
    datetime(2019, 1, 11, 19, 10, 20)
]

# Test the DST transition for Continental US
LOCAL_DATETIMES_DST = [
    datetime(2019, 3, 9, 12, 10, 20),
    datetime(2019, 3, 19, 12, 10, 20),
    datetime(2019, 11, 2, 12, 10, 20),
    datetime(2019, 11, 3, 12, 10, 20)
]

LOCAL_TIMESTAMPS = [pd.Timestamp(d) for d in LOCAL_DATETIMES]
LOCAL_TIMESTAMPS_DST = [pd.Timestamp(d) for d in LOCAL_DATETIMES_DST]

# Set up testing data
UTC = pytz.UTC

UTC_DATETIMES = [UTC.localize(d) for d in LOCAL_DATETIMES]
UTC_TIMESTAMPS = [UTC.localize(d) for d in LOCAL_TIMESTAMPS]

UTC_DATETIMES_DST = [UTC.localize(d, is_dst=True) for d in LOCAL_DATETIMES_DST]
UTC_TIMESTAMPS_DST = [UTC.localize(d, is_dst=True) for d in LOCAL_TIMESTAMPS_DST]

PST = pytz.timezone("US/Pacific")
CST = pytz.timezone("US/Central")
EST = pytz.timezone("US/Eastern")
GMT = pytz.timezone("GMT")
HKT = pytz.timezone("Asia/Hong_Kong")
JPT = pytz.timezone("Asia/Tokyo")
ACT = pytz.timezone("Australia/ACT")

TIMEZONES = [PST, CST, EST, GMT, HKT, JPT, ACT]

TZ_DATETIMES = {}
TZ_TIMESTAMPS = {}

TZ_DATETIMES_DST = {}
TZ_TIMESTAMPS_DST = {}

for TZ in TIMEZONES:
    TZ_DATETIMES[TZ.zone] = [TZ.localize(d) for d in LOCAL_DATETIMES]
    TZ_TIMESTAMPS[TZ.zone] = [d.tz_localize(TZ) for d in LOCAL_TIMESTAMPS]
    TZ_DATETIMES_DST[TZ.zone] = [d.astimezone(TZ) for d in UTC_DATETIMES_DST]
    TZ_TIMESTAMPS_DST[TZ.zone] = [d.tz_convert(TZ) for d in UTC_TIMESTAMPS_DST]

if os.name != 'nt':
    # no tzset on windows, run these tests on linux/mac only

    class TestTableLocalDateTime(object):
        """Test datetimes across configurations such as local time, timezone-aware,
        timezone-naive, and UTC implementations.
        """
        def setup_method(self):
            # To make sure that local times are not changed, set timezone to EST
            os.environ["TZ"] = "US/Eastern"
            time.tzset()

        def teardown_method(self):
            # go back to UTC at end of each test method, for consistency
            os.environ["TZ"] = "UTC"
            time.tzset()

        def test_table_should_assume_local_time(self):
            """If a datetime object has no `tzinfo`, it should be assumed to be in
            local time and not be converted at all.
            """
            data = {
                "a": LOCAL_DATETIMES
            }
            table = Table(data)
            assert table.view().to_dict()["a"] == LOCAL_DATETIMES

        def test_table_should_assume_local_time_numpy_datetime64(self):
            data = {
                "a": [np.datetime64(d) for d in LOCAL_DATETIMES]
            }
            table = Table(data)
            assert table.view().to_dict()["a"] == LOCAL_DATETIMES

        def test_table_should_assume_local_time_pandas_timestamp(self):
            data = {
                "a": LOCAL_TIMESTAMPS
            }

            # Timestamps are assumed to be in UTC by pandas
            table = Table(data)

            # Timestamps are read out in local time
            assert table.view().to_dict()["a"] == LOCAL_DATETIMES

        def test_table_should_assume_local_time_pandas_timestamp_df(self):
            data = pd.DataFrame({
                "a": LOCAL_TIMESTAMPS
            })

            # Timestamps are assumed to be in UTC by pandas
            table = Table(data)

            # Timestamps are read out in local time
            assert table.view().to_dict()["a"] == [
                datetime(2019, 1, 10, 19, 10, 20),
                datetime(2019, 1, 11, 6, 10, 20),
                datetime(2019, 1, 11, 14, 10, 20)
            ]

        def test_table_should_assume_local_time_dst(self):
            """If a datetime object has no `tzinfo`, it should be assumed to be in
            local time and not be converted at all.
            """
            data = {
                "a": LOCAL_DATETIMES_DST
            }
            table = Table(data)
            assert table.view().to_dict()["a"] == LOCAL_DATETIMES_DST

        def test_table_should_assume_local_time_numpy_datetime64_dst(self):
            data = {
                "a": [np.datetime64(d) for d in LOCAL_DATETIMES_DST]
            }
            table = Table(data)
            assert table.view().to_dict()["a"] == LOCAL_DATETIMES_DST

        def test_table_should_assume_local_time_pandas_timestamp_dst(self):
            data = {
                "a": LOCAL_TIMESTAMPS_DST
            }
            table = Table(data)
            assert table.view().to_dict()["a"] == LOCAL_DATETIMES_DST

        def test_table_should_assume_local_time_pandas_timestamp_dst_df(self):
            data = pd.DataFrame({
                "a": LOCAL_TIMESTAMPS_DST
            })
            table = Table(data)
            assert table.view().to_dict()["a"] == [
                datetime(2019, 3, 9, 7, 10, 20),
                datetime(2019, 3, 19, 8, 10, 20),
                datetime(2019, 11, 2, 8, 10, 20),
                datetime(2019, 11, 3, 7, 10, 20)
            ]


    class TestTableDateTimeUTCToLocal(object):

        def teardown_method(self):
            # Set timezone to UTC, always
            os.environ["TZ"] = "UTC"
            time.tzset()

        def test_table_should_convert_UTC_to_local_time_pytz_pacific(self):
            """If the datetime has `tzinfo` set, use it to convert the datetime to
            UTC. Make sure this works with both `pytz` and `dateutil` for
            `datetime` and `pandas.Timestamp`.
            """
            data = {
                "a": UTC_DATETIMES
            }
            table = Table(data)

            os.environ["TZ"] = "US/Pacific"
            time.tzset()

            # Should be in PST now
            assert table.view().to_dict() == {
                "a": [d.astimezone(PST).replace(tzinfo=None) for d in data["a"]]
            }

        def test_table_should_convert_UTC_to_local_time_pytz_central(self):
            data = {
                "a": UTC_DATETIMES
            }
            table = Table(data)

            os.environ["TZ"] = "US/Central"
            time.tzset()

            # Should be in CST now
            assert table.view().to_dict() == {
                "a": [d.astimezone(CST).replace(tzinfo=None) for d in data["a"]]
            }

        def test_table_should_convert_UTC_to_local_time_pytz_eastern(self):
            data = {
                "a": UTC_DATETIMES
            }
            table = Table(data)

            os.environ["TZ"] = "US/Eastern"
            time.tzset()

            # Should be in EST now
            assert table.view().to_dict() == {
                "a": [d.astimezone(EST).replace(tzinfo=None) for d in data["a"]]
            }

        def test_table_should_convert_UTC_to_local_time_pytz_GMT(self):
            data = {
                "a": UTC_DATETIMES
            }
            table = Table(data)

            os.environ["TZ"] = "GMT"
            time.tzset()

            # Should be in GMT now
            assert table.view().to_dict() == {
                "a": [d.astimezone(GMT).replace(tzinfo=None) for d in data["a"]]
            }

        def test_table_should_convert_UTC_to_local_time_pytz_HKT(self):
            data = {
                "a": UTC_DATETIMES
            }
            table = Table(data)

            os.environ["TZ"] = "Asia/Hong_Kong"
            time.tzset()

            assert table.view().to_dict() == {
                "a": [d.astimezone(HKT).replace(tzinfo=None) for d in data["a"]]
            }

        def test_table_should_convert_UTC_to_local_time_pytz_JPT(self):
            data = {
                "a": UTC_DATETIMES
            }
            table = Table(data)

            os.environ["TZ"] = "Asia/Tokyo"
            time.tzset()

            assert table.view().to_dict() == {
                "a": [d.astimezone(JPT).replace(tzinfo=None) for d in data["a"]]
            }

        def test_table_should_convert_UTC_to_local_time_pytz_ACT(self):
            data = {
                "a": UTC_DATETIMES
            }
            table = Table(data)

            os.environ["TZ"] = "Australia/Sydney"
            time.tzset()

            assert table.view().to_dict() == {
                "a": [d.astimezone(ACT).replace(tzinfo=None) for d in data["a"]]
            }

        def test_table_should_convert_UTC_to_local_time_dateutil_pacific(self):
            data = {
                "a": UTC_DATETIMES
            }
            table = Table(data)

            os.environ["TZ"] = "US/Pacific"
            time.tzset()

            # Should be in PST now
            assert table.view().to_dict() == {
                "a": [d.astimezone(PST).replace(tzinfo=None) for d in data["a"]]
            }

        def test_table_should_convert_UTC_to_local_time_dateutil_central(self):
            data = {
                "a": UTC_DATETIMES
            }
            table = Table(data)

            os.environ["TZ"] = "US/Central"
            time.tzset()

            # Should be in CST now
            assert table.view().to_dict() == {
                "a": [d.astimezone(CST).replace(tzinfo=None) for d in data["a"]]
            }

        def test_table_should_convert_UTC_to_local_time_dateutil_eastern(self):
            data = {
                "a": UTC_DATETIMES
            }
            table = Table(data)

            os.environ["TZ"] = "US/Eastern"
            time.tzset()

            # Should be in EST now
            assert table.view().to_dict() == {
                "a": [d.astimezone(EST).replace(tzinfo=None) for d in data["a"]]
            }

        def test_table_should_convert_UTC_to_local_time_dateutil_GMT(self):
            data = {
                "a": UTC_DATETIMES
            }
            table = Table(data)

            os.environ["TZ"] = "GMT"
            time.tzset()

            # Should be in GMT now
            assert table.view().to_dict() == {
                "a": [d.astimezone(GMT).replace(tzinfo=None) for d in data["a"]]
            }

        def test_table_should_convert_UTC_to_local_time_dateutil_pacific_DST(self):
            data = {
                "a": UTC_DATETIMES_DST
            }
            table = Table(data)

            os.environ["TZ"] = "US/Pacific"
            time.tzset()

            # Should be in PST now
            assert table.view().to_dict() == {
                "a": [d.replace(tzinfo=None) for d in TZ_DATETIMES_DST["US/Pacific"]]
            }

        def test_table_should_convert_UTC_to_local_time_dateutil_central_DST(self):
            data = {
                "a": UTC_DATETIMES_DST
            }
            table = Table(data)

            os.environ["TZ"] = "US/Central"
            time.tzset()

            # Should be in CST now
            assert table.view().to_dict() == {
                "a": [d.replace(tzinfo=None) for d in TZ_DATETIMES_DST["US/Central"]]
            }

        def test_table_should_convert_UTC_to_local_time_dateutil_eastern_DST(self):
            data = {
                "a": UTC_DATETIMES_DST
            }
            table = Table(data)

            os.environ["TZ"] = "US/Eastern"
            time.tzset()

            # Should be in EST now
            assert table.view().to_dict() == {
                "a": [d.replace(tzinfo=None) for d in TZ_DATETIMES_DST["US/Eastern"]]
            }

        def test_table_should_convert_UTC_to_local_time_dateutil_GMT_DST(self):
            data = {
                "a": UTC_DATETIMES_DST
            }
            table = Table(data)

            os.environ["TZ"] = "GMT"
            time.tzset()

            # Should be in GMT now
            assert table.view().to_dict() == {
                "a": [d.replace(tzinfo=None) for d in TZ_DATETIMES_DST["GMT"]]
            }

        def test_table_should_convert_UTC_to_local_time_dateutil_pacific_DST_timestamp(self):
            data = pd.DataFrame({
                "a": UTC_TIMESTAMPS_DST
            })
            table = Table(data)

            os.environ["TZ"] = "US/Pacific"
            time.tzset()

            # Should be in PST now
            assert table.view().to_dict()["a"] == [d.replace(tzinfo=None) for d in TZ_DATETIMES_DST["US/Pacific"]]

        def test_table_should_convert_UTC_to_local_time_dateutil_central_DST_timestamp(self):
            data = pd.DataFrame({
                "a": UTC_TIMESTAMPS_DST
            })
            table = Table(data)

            os.environ["TZ"] = "US/Central"
            time.tzset()

            # Should be in CST now
            assert table.view().to_dict()["a"] == [d.replace(tzinfo=None) for d in TZ_DATETIMES_DST["US/Central"]]

        def test_table_should_convert_UTC_to_local_time_dateutil_eastern_DST_timestamp(self):
            data = pd.DataFrame({
                "a": UTC_TIMESTAMPS_DST
            })
            table = Table(data)

            os.environ["TZ"] = "US/Eastern"
            time.tzset()

            # Should be in EST now
            assert table.view().to_dict()["a"] == [d.replace(tzinfo=None) for d in TZ_DATETIMES_DST["US/Eastern"]]

        def test_table_should_convert_UTC_to_local_time_dateutil_GMT_DST_timestamp(self):
            data = pd.DataFrame({
                "a": UTC_TIMESTAMPS_DST
            })
            table = Table(data)

            os.environ["TZ"] = "GMT"
            time.tzset()

            # Should be in GMT now
            assert table.view().to_dict()["a"] == [d.replace(tzinfo=None) for d in TZ_DATETIMES_DST["GMT"]]

        def test_table_should_convert_UTC_to_local_time_dateutil_HKT(self):
            data = {
                "a": UTC_DATETIMES
            }
            table = Table(data)

            os.environ["TZ"] = "Asia/Hong_Kong"
            time.tzset()

            assert table.view().to_dict() == {
                "a": [d.astimezone(HKT).replace(tzinfo=None) for d in data["a"]]
            }

        def test_table_should_convert_UTC_to_local_time_dateutil_JPT(self):
            data = {
                "a": UTC_DATETIMES
            }
            table = Table(data)

            os.environ["TZ"] = "Asia/Tokyo"
            time.tzset()

            assert table.view().to_dict() == {
                "a": [d.astimezone(JPT).replace(tzinfo=None) for d in data["a"]]
            }

        def test_table_should_convert_UTC_to_local_time_dateutil_ACT(self):
            data = {
                "a": UTC_DATETIMES
            }
            table = Table(data)

            os.environ["TZ"] = "Australia/Sydney"
            time.tzset()

            ACT = tz.gettz("Australia/Sydney")

            assert table.view().to_dict() == {
                "a": [d.astimezone(ACT).replace(tzinfo=None) for d in data["a"]]
            }

        def test_table_should_convert_UTC_to_local_time_pytz_pacific_timestamp(self):
            data = pd.DataFrame({
                "a": UTC_TIMESTAMPS
            })
            table = Table(data)

            os.environ["TZ"] = "US/Pacific"
            time.tzset()

            # Should be in PST now
            assert table.view().to_dict()["a"] == [d.astimezone(PST).replace(tzinfo=None) for d in data["a"]]

        def test_table_should_convert_UTC_to_local_time_pytz_central_timestamp(self):
            data = pd.DataFrame({
                "a": UTC_TIMESTAMPS
            })
            table = Table(data)

            os.environ["TZ"] = "US/Central"
            time.tzset()

            # Should be in CST now
            assert table.view().to_dict()["a"] == [d.astimezone(CST).replace(tzinfo=None) for d in data["a"]]

        def test_table_should_convert_UTC_to_local_time_pytz_eastern_timestamp(self):
            data = pd.DataFrame({
                "a": UTC_TIMESTAMPS
            })
            table = Table(data)

            os.environ["TZ"] = "US/Eastern"
            time.tzset()

            # Should be in EST now
            assert table.view().to_dict()["a"] == [d.astimezone(EST).replace(tzinfo=None) for d in data["a"]]

        def test_table_should_convert_UTC_to_local_time_pytz_GMT_timestamp(self):
            data = pd.DataFrame({
                "a": UTC_TIMESTAMPS
            })
            table = Table(data)

            os.environ["TZ"] = "GMT"
            time.tzset()

            # Should be in GMT now
            assert table.view().to_dict()["a"] == [d.astimezone(GMT).replace(tzinfo=None) for d in data["a"]]

        def test_table_should_convert_UTC_to_local_time_pytz_HKT_timestamp(self):
            data = pd.DataFrame({
                "a": UTC_TIMESTAMPS
            })
            table = Table(data)

            os.environ["TZ"] = "Asia/Hong_Kong"
            time.tzset()

            assert table.view().to_dict()["a"] == [d.astimezone(HKT).replace(tzinfo=None) for d in data["a"]]

        def test_table_should_convert_UTC_to_local_time_pytz_JPT_timestamp(self):
            data = pd.DataFrame({
                "a": UTC_TIMESTAMPS
            })
            table = Table(data)

            os.environ["TZ"] = "Asia/Tokyo"
            time.tzset()

            assert table.view().to_dict()["a"] == [d.astimezone(JPT).replace(tzinfo=None) for d in data["a"]]

        def test_table_should_convert_UTC_to_local_time_pytz_ACT_timestamp(self):
            data = pd.DataFrame({
                "a": UTC_TIMESTAMPS
            })
            table = Table(data)

            os.environ["TZ"] = "Australia/Sydney"
            time.tzset()

            assert table.view().to_dict()["a"] == [d.astimezone(ACT).replace(tzinfo=None) for d in data["a"]]

        def test_table_should_convert_UTC_to_local_time_dateutil_pacific_timestamp(self):
            data = pd.DataFrame({
                "a": UTC_TIMESTAMPS
            })
            table = Table(data)

            os.environ["TZ"] = "US/Pacific"
            time.tzset()

            # Should be in PST now
            assert table.view().to_dict()["a"] == [d.astimezone(PST).replace(tzinfo=None) for d in data["a"]]

        def test_table_should_convert_UTC_to_local_time_dateutil_central_timestamp(self):
            data = pd.DataFrame({
                "a": UTC_TIMESTAMPS
            })
            table = Table(data)

            os.environ["TZ"] = "US/Central"
            time.tzset()

            CST = tz.gettz("US/Central")

            # Should be in CST now
            assert table.view().to_dict()["a"] == [d.astimezone(CST).replace(tzinfo=None) for d in data["a"]]

        def test_table_should_convert_UTC_to_local_time_dateutil_eastern_timestamp(self):
            data = pd.DataFrame({
                "a": UTC_TIMESTAMPS
            })
            table = Table(data)

            os.environ["TZ"] = "US/Eastern"
            time.tzset()

            # Should be in EST now
            assert table.view().to_dict()["a"] == [d.astimezone(EST).replace(tzinfo=None) for d in data["a"]]

        def test_table_should_convert_UTC_to_local_time_dateutil_GMT_timestamp(self):
            data = pd.DataFrame({
                "a": UTC_TIMESTAMPS
            })
            table = Table(data)

            os.environ["TZ"] = "GMT"
            time.tzset()

            GMT = tz.gettz("GMT")

            # Should be in GMT now
            assert table.view().to_dict()["a"] == [d.astimezone(GMT).replace(tzinfo=None) for d in data["a"]]

        def test_table_should_convert_UTC_to_local_time_dateutil_HKT_timestamp(self):
            data = pd.DataFrame({
                "a": UTC_TIMESTAMPS
            })
            table = Table(data)

            os.environ["TZ"] = "Asia/Hong_Kong"
            time.tzset()

            assert table.view().to_dict()["a"] == [d.astimezone(HKT).replace(tzinfo=None) for d in data["a"]]

        def test_table_should_convert_UTC_to_local_time_dateutil_JPT_timestamp(self):
            data = pd.DataFrame({
                "a": UTC_TIMESTAMPS
            })
            table = Table(data)

            os.environ["TZ"] = "Asia/Tokyo"
            time.tzset()

            assert table.view().to_dict()["a"] == [d.astimezone(JPT).replace(tzinfo=None) for d in data["a"]]

        def test_table_should_convert_UTC_to_local_time_dateutil_ACT_timestamp(self):
            data = pd.DataFrame({
                "a": UTC_TIMESTAMPS
            })
            table = Table(data)

            os.environ["TZ"] = "Australia/Sydney"
            time.tzset()

            assert table.view().to_dict()["a"] == [d.astimezone(ACT).replace(tzinfo=None) for d in data["a"]]


    class TestTableDateTimeArbitaryToLocal(object):

        def teardown_method(self):
            # Set timezone to UTC, always
            os.environ["TZ"] = "UTC"
            time.tzset()

        def test_table_should_convert_PST_to_local_time_pytz_central(self):
            data = {
                "a": TZ_DATETIMES["US/Pacific"]
            }
            table = Table(data)

            os.environ["TZ"] = "US/Central"
            time.tzset()

            # Should be in CST now
            assert table.view().to_dict() == {
                "a": [d.astimezone(CST).replace(tzinfo=None) for d in data["a"]]
            }

        def test_table_should_convert_CST_to_local_time_pytz_eastern(self):
            data = {
                "a": TZ_DATETIMES["US/Central"]
            }
            table = Table(data)

            os.environ["TZ"] = "US/Eastern"
            time.tzset()

            # Should be in EST now
            assert table.view().to_dict() == {
                "a": [d.astimezone(EST).replace(tzinfo=None) for d in data["a"]]
            }

        def test_table_should_convert_EST_to_local_time_pytz_GMT(self):
            data = {
                "a": TZ_DATETIMES["US/Eastern"]
            }
            table = Table(data)

            os.environ["TZ"] = "GMT"
            time.tzset()

            # Should be in GMT now
            assert table.view().to_dict() == {
                "a": [d.astimezone(GMT).replace(tzinfo=None) for d in data["a"]]
            }

        def test_table_should_convert_GMT_to_local_time_pytz_HKT(self):
            data = {
                "a": TZ_DATETIMES["GMT"]
            }
            table = Table(data)

            os.environ["TZ"] = "Asia/Hong_Kong"
            time.tzset()

            assert table.view().to_dict() == {
                "a": [d.astimezone(HKT).replace(tzinfo=None) for d in data["a"]]
            }

        def test_table_should_convert_HKT_to_local_time_pytz_JPT(self):
            data = {
                "a": TZ_DATETIMES["Asia/Hong_Kong"]
            }
            table = Table(data)

            os.environ["TZ"] = "Asia/Tokyo"
            time.tzset()

            assert table.view().to_dict() == {
                "a": [d.astimezone(JPT).replace(tzinfo=None) for d in data["a"]]
            }

        def test_table_should_convert_JPT_to_local_time_pytz_ACT(self):
            data = {
                "a": TZ_DATETIMES["Asia/Tokyo"]
            }
            table = Table(data)

            os.environ["TZ"] = "Australia/Sydney"
            time.tzset()

            assert table.view().to_dict() == {
                "a": [d.astimezone(ACT).replace(tzinfo=None) for d in data["a"]]
            }

        def test_table_should_convert_PST_to_local_time_dateutil_central(self):
            data = {
                "a": TZ_DATETIMES["US/Pacific"]
            }
            table = Table(data)

            os.environ["TZ"] = "US/Central"
            time.tzset()

            # Should be in CST now
            assert table.view().to_dict() == {
                "a": [d.astimezone(CST).replace(tzinfo=None) for d in data["a"]]
            }

        def test_table_should_convert_CST_to_local_time_dateutil_eastern(self):
            data = {
                "a": TZ_DATETIMES["US/Central"]
            }
            table = Table(data)

            os.environ["TZ"] = "US/Eastern"
            time.tzset()

            # Should be in EST now
            assert table.view().to_dict() == {
                "a": [d.astimezone(EST).replace(tzinfo=None) for d in data["a"]]
            }

        def test_table_should_convert_EST_to_local_time_dateutil_GMT(self):
            data = {
                "a": TZ_DATETIMES["US/Eastern"]
            }
            table = Table(data)

            os.environ["TZ"] = "GMT"
            time.tzset()

            # Should be in GMT now
            assert table.view().to_dict() == {
                "a": [d.astimezone(GMT).replace(tzinfo=None) for d in data["a"]]
            }

        def test_table_should_convert_GMT_to_local_time_dateutil_HKT(self):
            data = {
                "a": TZ_DATETIMES["GMT"]
            }
            table = Table(data)

            os.environ["TZ"] = "Asia/Hong_Kong"
            time.tzset()

            assert table.view().to_dict() == {
                "a": [d.astimezone(HKT).replace(tzinfo=None) for d in data["a"]]
            }

        def test_table_should_convert_HKT_to_local_time_dateutil_JPT(self):
            data = {
                "a": TZ_DATETIMES["Asia/Hong_Kong"]
            }
            table = Table(data)

            os.environ["TZ"] = "Asia/Tokyo"
            time.tzset()

            assert table.view().to_dict() == {
                "a": [d.astimezone(JPT).replace(tzinfo=None) for d in data["a"]]
            }

        def test_table_should_convert_JPT_to_local_time_dateutil_ACT(self):
            data = {
                "a": TZ_DATETIMES["Asia/Tokyo"]
            }
            table = Table(data)

            os.environ["TZ"] = "Australia/Sydney"
            time.tzset()

            assert table.view().to_dict() == {
                "a": [d.astimezone(ACT).replace(tzinfo=None) for d in data["a"]]
            }

        def test_table_should_convert_PST_to_local_time_pytz_central_timestamp(self):
            data = {
                "a": TZ_TIMESTAMPS["US/Pacific"]
            }
            table = Table(pd.DataFrame(data))

            os.environ["TZ"] = "US/Central"
            time.tzset()

            # Should be in CST now
            assert table.view().to_dict()["a"] == [d.astimezone(CST).replace(tzinfo=None) for d in data["a"]]

        def test_table_should_convert_CST_to_local_time_pytz_eastern_timestamp(self):
            data = {
                "a": TZ_TIMESTAMPS["US/Central"]
            }
            table = Table(pd.DataFrame(data))

            os.environ["TZ"] = "US/Eastern"
            time.tzset()

            # Should be in EST now
            assert table.view().to_dict()["a"] == [d.astimezone(EST).replace(tzinfo=None) for d in data["a"]]

        def test_table_should_convert_EST_to_local_time_pytz_GMT_timestamp(self):
            data = {
                "a": TZ_TIMESTAMPS["US/Eastern"]
            }
            table = Table(pd.DataFrame(data))

            os.environ["TZ"] = "GMT"
            time.tzset()

            # Should be in GMT now
            assert table.view().to_dict()["a"] == [d.astimezone(GMT).replace(tzinfo=None) for d in data["a"]]

        def test_table_should_convert_GMT_to_local_time_pytz_HKT_timestamp(self):
            data = {
                "a": TZ_TIMESTAMPS["GMT"]
            }
            table = Table(pd.DataFrame(data))

            os.environ["TZ"] = "Asia/Hong_Kong"
            time.tzset()

            assert table.view().to_dict()["a"] == [d.astimezone(HKT).replace(tzinfo=None) for d in data["a"]]

        def test_table_should_convert_HKT_to_local_time_pytz_JPT_timestamp(self):
            data = {
                "a": TZ_TIMESTAMPS["Asia/Hong_Kong"]
            }
            table = Table(pd.DataFrame(data))

            os.environ["TZ"] = "Asia/Tokyo"
            time.tzset()

            assert table.view().to_dict()["a"] == [d.astimezone(JPT).replace(tzinfo=None) for d in data["a"]]

        def test_table_should_convert_JPT_to_local_time_pytz_ACT_timestamp(self):
            data = {
                "a": TZ_TIMESTAMPS["Asia/Tokyo"]
            }
            table = Table(pd.DataFrame(data))

            os.environ["TZ"] = "Australia/Sydney"
            time.tzset()

            assert table.view().to_dict()["a"] == [d.astimezone(ACT).replace(tzinfo=None) for d in data["a"]]

        def test_table_should_convert_PST_to_local_time_dateutil_central_timestamp(self):
            data = {
                "a": TZ_TIMESTAMPS["US/Pacific"]
            }
            table = Table(pd.DataFrame(data))

            os.environ["TZ"] = "US/Central"
            time.tzset()

            # Should be in CST now
            assert table.view().to_dict()["a"] == [d.astimezone(CST).replace(tzinfo=None) for d in data["a"]]

        def test_table_should_convert_CST_to_local_time_dateutil_eastern_timestamp(self):
            data = {
                "a": TZ_TIMESTAMPS["US/Central"]
            }
            table = Table(pd.DataFrame(data))

            os.environ["TZ"] = "US/Eastern"
            time.tzset()

            # Should be in EST now
            assert table.view().to_dict()["a"] == [d.astimezone(EST).replace(tzinfo=None) for d in data["a"]]

        def test_table_should_convert_EST_to_local_time_dateutil_GMT_timestamp(self):
            data = {
                "a": TZ_TIMESTAMPS["US/Eastern"]
            }
            table = Table(pd.DataFrame(data))

            os.environ["TZ"] = "GMT"
            time.tzset()

            # Should be in GMT now
            assert table.view().to_dict()["a"] == [d.astimezone(GMT).replace(tzinfo=None) for d in data["a"]]

        def test_table_should_convert_GMT_to_local_time_dateutil_HKT_timestamp(self):
            data = {
                "a": TZ_TIMESTAMPS["GMT"]
            }
            table = Table(pd.DataFrame(data))

            os.environ["TZ"] = "Asia/Hong_Kong"
            time.tzset()

            assert table.view().to_dict()["a"] == [d.astimezone(HKT).replace(tzinfo=None) for d in data["a"]]

        def test_table_should_convert_HKT_to_local_time_dateutil_JPT_timestamp(self):
            data = {
                "a": TZ_TIMESTAMPS["Asia/Hong_Kong"]
            }
            table = Table(pd.DataFrame(data))

            os.environ["TZ"] = "Asia/Tokyo"
            time.tzset()

            assert table.view().to_dict()["a"] == [d.astimezone(JPT).replace(tzinfo=None) for d in data["a"]]

        def test_table_should_convert_JPT_to_local_time_dateutil_ACT_timestamp(self):
            data = {
                "a": TZ_TIMESTAMPS["Asia/Tokyo"]
            }
            table = Table(pd.DataFrame(data))

            os.environ["TZ"] = "Australia/Sydney"
            time.tzset()

            assert table.view().to_dict()["a"] == [d.astimezone(ACT).replace(tzinfo=None) for d in data["a"]]



import math
import numpy as np
import scipy as sp
from .utils import _CALCULATIONS_GRAPHVIZSHAPE
from ..base import Node


def unary(node, name, lam):
    return node._gennode(name, lam, [node], node._trace, graphvizshape=_CALCULATIONS_GRAPHVIZSHAPE)


def binary(node1, other, name, lam):
    if isinstance(node1._self_reference, Node):
        return node1._gennode(name, lam, [node1._self_reference, other], node1._trace or other._trace, graphvizshape=_CALCULATIONS_GRAPHVIZSHAPE)
    return node1._gennode(name, lam, [node1, other], node1._trace or other._trace, graphvizshape=_CALCULATIONS_GRAPHVIZSHAPE)


def n_nary(node, others, name, lam):
    if isinstance(node._self_reference, Node):
        return node._gennode(name,
                             lam,
                             [node._self_reference] + others,
                             node._trace or any(other._trace for other in others),
                             graphvizshape=_CALCULATIONS_GRAPHVIZSHAPE)
    return node._gennode(name,
                         lam,
                         [node] + others,
                         node._trace or any(other._trace for other in others),
                         graphvizshape=_CALCULATIONS_GRAPHVIZSHAPE)

########################
# Arithmetic Operators #
########################


def Add(self, other):
    other = self._tonode(other)
    return binary(self, other, self._name_no_id() + '+' + other._name_no_id(), (lambda x, y: x.value() + y.value()))


def Sub(self, other):
    other = self._tonode(other)
    return binary(self, other, self._name_no_id() + '-' + other._name_no_id(), (lambda x, y: x.value() - y.value()))


def Mult(self, other):
    other = self._tonode(other)
    return binary(self, other, self._name_no_id() + '*' + other._name_no_id(), (lambda x, y: x.value() * y.value()))


def Div(self, other):
    other = self._tonode(other)
    return binary(self, other, self._name_no_id() + '/' + other._name_no_id(), (lambda x, y: x.value() / y.value()))


def Pow(self, other):
    other = self._tonode(other)
    return binary(self, other, self._name_no_id() + '^' + other._name_no_id(), (lambda x, y: x.value() ** y.value()))


def Mod(self, other):
    other = self._tonode(other)
    return binary(self, other, self._name_no_id() + '%' + other._name_no_id(), (lambda x, y: x.value() % y.value()))


def Negate(self):
    return unary(self, '(-' + self._name_no_id() + ')', (lambda x: -self.value()))


def Invert(self):
    return unary(self, '1/' + self._name_no_id(), (lambda x: 1 / self.value()))


def Sum(self, *others):
    others_nodes = []
    for other in others:
        others_nodes.append(self._tonode(other))

    return n_nary(self,
                  others_nodes,
                  'Sum(' + self._name_no_id() + ',' + ','.join(other._name_no_id() for other in others_nodes) + ')',
                  (lambda *args: sum(x.value() for x in args)))


def Average(self, *others):
    others_nodes = []
    for other in others:
        others_nodes.append(self._tonode(other))

    return n_nary(self,
                  others_nodes,
                  'Average(' + self._name_no_id() + ',' + ','.join(other._name_no_id() for other in others_nodes) + ')',
                  (lambda *args: sum(x.value() for x in args) / len(args)))


#####################
# Logical Operators #
#####################
def Or(self, other):
    other = self._tonode(other)
    return binary(self, other, self._name_no_id() + '||' + other._name_no_id(), (lambda x, y: x.value() or y.value()))


def And(self, other):
    other = self._tonode(other)
    return binary(self, other, self._name_no_id() + '&&' + other._name_no_id(), (lambda x, y: x.value() and y.value()))


def Not(self):
    return unary(self, '!' + self._name_no_id(), (lambda x: not x.value()))


##########################
# Mathematical Functions #
##########################
def Sin(self):
    return unary(self, 'sin(' + self._name_no_id() + ')', (lambda x: math.sin(self.value())))


def Cos(self):
    return unary(self, 'cos(' + self._name_no_id() + ')', (lambda x: math.cos(self.value())))


def Tan(self):
    return unary(self, 'tan(' + self._name_no_id() + ')', (lambda x: math.tan(self.value())))


def Arcsin(self):
    return unary(self, 'arcsin(' + self._name_no_id() + ')', (lambda x: math.asin(self.value())))


def Arccos(self):
    return unary(self, 'arccos(' + self._name_no_id() + ')', (lambda x: math.acos(self.value())))


def Arctan(self):
    return unary(self, 'arctan(' + self._name_no_id() + ')', (lambda x: math.atan(self.value())))


def Abs(self):
    return unary(self, '||' + self._name_no_id() + '||', (lambda x: abs(self.value())))


def Sqrt(self):
    return unary(self, 'sqrt(' + str(self._name_no_id()) + ')', (lambda x: math.sqrt(self.value())))


def Log(self):
    return unary(self, 'log(' + str(self._name_no_id()) + ')', (lambda x: math.log(self.value())))


def Exp(self):
    return unary(self, 'exp(' + str(self._name_no_id()) + ')', (lambda x: math.exp(self.value())))


def Erf(self):
    return unary(self, 'erf(' + str(self._name_no_id()) + ')', (lambda x: math.erf(self.value())))


def Floor(self):
    return unary(self, 'floor(' + str(self._name_no_id()) + ')', (lambda x: math.floor(self.value())))


def Ceil(self):
    return unary(self, 'ceil(' + str(self._name_no_id()) + ')', (lambda x: math.ceil(self.value())))


def Round(self, ndigits=0):
    return unary(self, 'round(' + str(self._name_no_id()) + ')', (lambda x: round(self.value(), ndigits=ndigits)))


##############
# Converters #
##############
def Float(self):
    return unary(self, 'float(' + str(self._name_no_id()) + ')', (lambda x: float(self.value())))


def Int(self):
    return unary(self, 'int(' + str(self._name_no_id()) + ')', (lambda x: int(self.value())))


def Bool(self):
    return unary(self, 'bool(' + str(self._name_no_id()) + ')', (lambda x: bool(self.value())))


def __Bool__(self):
    if self.value() is None:
        return False
    return bool(self.value())


def Str(self):
    return unary(self, 'str(' + str(self._name_no_id()) + ')', (lambda x: str(self.value())))


###################
# Python Builtins #
###################
def Len(self):
    return unary(self, 'len(' + str(self._name_no_id()) + ')', (lambda x: len(self.value())))


###################
# Numpy Functions #
###################
def __array_ufunc__(self, ufunc, method, *inputs, **kwargs):
    if ufunc == np.add:
        if isinstance(inputs[0], Node):
            return inputs[0].__add__(inputs[1])
        else:
            return inputs[1].__add__(inputs[0])
    elif ufunc == np.subtract:
        if isinstance(inputs[0], Node):
            return inputs[0].__sub__(inputs[1])
        else:
            return inputs[1].__sub__(inputs[0])
    elif ufunc == np.multiply:
        if isinstance(inputs[0], Node):
            return inputs[0].__mul__(inputs[1])
        else:
            return inputs[1].__mul__(inputs[0])
    elif ufunc == np.divide:
        if isinstance(inputs[0], Node):
            return inputs[0].__truediv__(inputs[1])
        else:
            return inputs[1].__truediv__(inputs[0])
    elif ufunc == np.sin:
        return inputs[0].sin()
    elif ufunc == np.cos:
        return inputs[0].cos()
    elif ufunc == np.tan:
        return inputs[0].tan()
    elif ufunc == np.arcsin:
        return inputs[0].asin()
    elif ufunc == np.arccos:
        return inputs[0].acos()
    elif ufunc == np.arctan:
        return inputs[0].atan()
    elif ufunc == np.exp:
        return inputs[0].exp()
    elif ufunc == sp.special.erf:
        return inputs[0].erf()
    raise NotImplementedError('Not Implemented!')


def __array_function__(self, func, method, *inputs, **kwargs):
    if func == np.lib.scimath.log:
        return inputs[0][0].log()
    elif func == np.lib.scimath.sqrt:
        return inputs[0][0].sqrt()
    raise NotImplementedError('Not Implemented!')


###############
# Comparators #
###############
def Equal(self, other):
    if isinstance(other, Node) and super(Node, self).__eq__(other):
        return True
    other = self._tonode(other)
    return binary(self, other, self._name_no_id() + '==' + other._name_no_id(), (lambda x, y: x.value() == y.value()))


def NotEqual(self, other):
    if isinstance(other, Node) and super(Node, self).__eq__(other):
        return False
    other = self._tonode(other)
    return binary(self, other, self._name_no_id() + '!=' + other._name_no_id(), (lambda x, y: x.value() != y.value()))


def Ge(self, other):
    if isinstance(other, Node) and super(Node, self).__eq__(other):
        return True
    other = self._tonode(other)
    return binary(self, other, self._name_no_id() + '>=' + other._name_no_id(), (lambda x, y: x.value() >= y.value()))


def Gt(self, other):
    if isinstance(other, Node) and super(Node, self).__eq__(other):
        return False
    other = self._tonode(other)
    return binary(self, other, self._name_no_id() + '>' + other._name_no_id(), (lambda x, y: x.value() > y.value()))


def Le(self, other):
    if isinstance(other, Node) and super(Node, self).__eq__(other):
        return True
    other = self._tonode(other)
    return binary(self, other, self._name_no_id() + '<=' + other._name_no_id(), (lambda x, y: x.value() <= y.value()))


def Lt(self, other):
    if isinstance(other, Node) and super(Node, self).__eq__(other):
        return False
    other = self._tonode(other)
    return binary(self, other, self._name_no_id() + '<' + other._name_no_id(), (lambda x, y: x.value() < y.value()))


########################
# Arithmetic Operators #
########################
Node.__add__ = Add
Node.__radd__ = Add
Node.__sub__ = Sub
Node.__rsub__ = Sub
Node.__mul__ = Mult
Node.__rmul__ = Mult
# Node.__matmul__ = MatMult
# Node.__rmatmul__ = MatMult
Node.__div__ = Div
Node.__rdiv__ = Div
# Node.__divmod__ = DivMod
# Node.__rdivmod__ = DivMod
Node.__truediv__ = Div
Node.__rtruediv__ = Div
Node.__floordiv__ = Div
# Node.__lshift__ = LShift
# Node.__rlshift__ = LShift
# Node.__rshift__ = RShift
# Node.__rrshift__ = RShift

Node.__pow__ = Pow
Node.__rpow__ = Pow
Node.__mod__ = Mod
Node.__rmod__ = Mod

Node.sum = Sum
Node.average = Average
Node.invert = Invert

#####################
# Logical Operators #
#####################
Node.__and__ = And
Node.__rand__ = And
Node.__or__ = Or
Node.__ror__ = Or
# Node.__xor__ = Xor
# Node.__rxor__ = Xor
Node.__invert__ = Not

##############
# Converters #
##############
Node.int = Int
Node.float = Float
Node.__bool__ = __Bool__

###############
# Comparators #
###############
Node.__lt__ = Lt
Node.__le__ = Le
Node.__gt__ = Gt
Node.__ge__ = Ge
Node.__eq__ = Equal
Node.__ne__ = NotEqual
Node.__neg__ = Negate
# Node.__pos__ = Pos
Node.__nonzero__ = Bool  # Py2 compat

###################
# Python Builtins #
###################
Node.__len__ = Len
# Node.__round__ = Len
# Node.__trunc__ = Len
# Node.__floor__ = Len
# Node.__ceil__ = Len

###################
# Numpy Functions #
###################
Node.__array_ufunc__ = __array_ufunc__
Node.__array_function__ = __array_function__

##########################
# Mathematical Functions #
##########################
Node.log = Log
Node.sin = Sin
Node.cos = Cos
Node.tan = Tan
Node.asin = Arcsin
Node.acos = Arccos
Node.atan = Arctan
Node.abs = Abs
Node.sqrt = Sqrt
Node.exp = Exp
Node.erf = Erf
Node.floor = Floor
Node.ceil = Ceil
Node.round = Round



from .base import StreamNone
from gevent import spawn


def run(target, timeout=1):
    '''Helper for running a thread

    Args:
        target (function): function to run on a thread
        timeout (int): how long to wait for target to return
    Returns:
        data: result of the function
    '''
    last = None
    done = False

    while not done:
        g = spawn(target)
        g.join(timeout)

        while not g.successful():
            yield StreamNone(last)

        last = g.value
        if last is None:
            done = True
        else:
            yield last



from traitlets import HasTraits, validate, observe, TraitError, Unicode, Dict


class Node(HasTraits):
    name = Unicode()
    label = Unicode()
    shape = Unicode(default_value="rect")  # rect, circle, ellipse, diamond
    tooltip = Unicode(default_value="")
    attrs = Dict()

    _graph = None

    def __init__(self, name, label="", shape="rect", tooltip="", **attrs):
        super(Node, self).__init__(self, name=name, label=label or name, shape=shape, tooltip=tooltip or label or name, attrs=attrs or {})

    def _setGraph(self, g):
        self._graph = g

    def _notify_change(self, attr, value):
        if not self._graph:
            return
        self._graph._notify_change(self, attr, value)

    @observe('name')
    def _observe_name(self, change):
        self._notify_change('name', change['new'])

    @observe('label')
    def _observe_label(self, change):
        self._notify_change('label', change['new'])

    @observe('shape')
    def _observe_shape(self, change):
        self._notify_change('shape', change['new'])

    @observe('tooltip')
    def _observe_tooltip(self, change):
        self._notify_change('tooltip', change['new'])

    @observe('attrs')
    def _observe_attrs(self, change):
        self._notify_change('attrs', change['new'])

    @validate("shape")
    def _validate_shape(self, proposal):
        if proposal.value not in ("rect", "circle", "ellipse", "diamond"):
            raise TraitError('Shape must be in ("rect", "circle", "ellipse", "diamond"), got: {}'.format(proposal))
        return proposal.value

    def to_dict(self):
        ret = {}
        ret["name"] = self.name
        ret["attrs"] = self.attrs
        ret["attrs"]["label"] = self.label
        ret["attrs"]["shape"] = self.shape
        ret["attrs"]["tooltip"] = self.tooltip or self.label
        return ret



from .node import Node
from traitlets import HasTraits, TraitError, observe, validate, Unicode, Instance, Float, Dict


class Edge(HasTraits):
    v = Instance(Node)
    w = Instance(Node)
    label = Unicode()
    attrs = Dict()
    tooltip = Unicode(default_value="")
    labelpos = Unicode(default_value="r")
    labeloffset = Float()
    arrowhead = Unicode(default_value="vee")

    _graph = None

    def __init__(self, v, w, label="", labelpos="r", labeloffset=0.0, tooltip="", **attrs):
        super(Edge, self).__init__(v=v, w=w, label=label, labelpos=labelpos, labeloffset=labeloffset, tooltip=tooltip or label or "{}->{}".format(v, w), attrs=attrs or {})

    def _setGraph(self, g):
        self._graph = g

    def _notify_change(self, attr, value):
        if not self._graph:
            return
        self._graph._notify_change(self, attr, value)

    @observe('v')
    def _observe_v(self, change):
        self._notify_change('v', change['new'])

    @observe('w')
    def _observe_w(self, change):
        self._notify_change('w', change['new'])

    @observe('label')
    def _observe_label(self, change):
        self._notify_change('label', change['new'])

    @observe('labelpos')
    def _observe_labelpos(self, change):
        self._notify_change('labelpos', change['new'])

    @observe('labeloffset')
    def _observe_labeloffset(self, change):
        self._notify_change('labeloffset', change['new'])

    @observe('tooltip')
    def _observe_tooltip(self, change):
        self._notify_change('tooltip', change['new'])

    @observe('attrs')
    def _observe_attrs(self, change):
        self._notify_change('attrs', change['new'])

    @validate("labelpos")
    def _validate_lablepos(self, proposal):
        if proposal.value not in ('l', 'c', 'r'):
            raise TraitError("lablepos must be in (l, c, r)")
        return proposal.value

    def to_dict(self):
        ret = {}
        ret["v"] = self.v.to_dict()
        ret["w"] = self.w.to_dict()
        ret["attrs"] = self.attrs
        ret["attrs"]["label"] = self.label
        ret["attrs"]["labelpos"] = self.labelpos
        ret["attrs"]["labeloffset"] = self.labeloffset
        ret["attrs"]["tooltip"] = self.tooltip
        ret["attrs"]["arrowhead"] = self.arrowhead
        return ret



import tributary.lazy as tl
import tributary.streaming as ts

from sympy.utilities.lambdify import lambdify
from sympy.parsing.sympy_parser import parse_expr, standard_transformations as _st, implicit_multiplication_application as _ima
from sympy import init_printing, dotprint, preorder_traversal
from graphviz import Source


init_printing(use_unicode=True)


def parse_expression(expr):
    '''Parse string as sympy expression
    Args:
        expr (string): string to convert to sympy expression
    '''
    return parse_expr(expr, transformations=(_st + (_ima,)))


def graphviz(expr):
    '''Plot sympy expression tree using graphviz
    Args:
        expr (sympy expression)
    '''

    return Source(dotprint(expr))


def traversal(expr):
    '''Traverse sympy expression tree
    Args:
        expr (sympy expression)
    '''

    return list(preorder_traversal(expr))


def symbols(expr):
    '''Get symbols used in sympy expression
    Args:
        expr (sympy expression)
    '''
    return expr.free_symbols


def construct_lazy(expr, modules=None):
    '''Construct Lazy tributary class from sympy expression

    Args:
        expr (sympy expression): A Sympy expression
        modules (list): a list of modules to use for sympy's lambdify function
    Returns:
        tributary.lazy.LazyGraph
    '''
    syms = list(symbols(expr))
    names = [s.name for s in syms]
    modules = modules or ["scipy", "numpy"]

    class Lazy(tl.LazyGraph):
        def __init__(self, **kwargs):
            for n in names:
                setattr(self, n, self.node(name=n, value=kwargs.get(n, None)))

            self._nodes = [getattr(self, n) for n in names]
            self._function = lambdify(syms, expr, modules=modules)
            self._expr = expr

        @tl.node
        def evaluate(self):
            return self._function(*self._nodes)

    return Lazy


def construct_streaming(expr, modules=None):
    '''Construct streaming tributary class from sympy expression

    Args:
        expr (sympy expression): A Sympy expression
        modules (list): a list of modules to use for sympy's lambdify function
    Returns:

    '''
    syms = list(symbols(expr))
    names = [s.name for s in syms]
    modules = modules or ["scipy", "numpy"]
    # modules = modules or  ["math", "mpmath", "sympy"]

    class Streaming(ts.StreamingGraph):
        def __init__(self, **kwargs):
            self._kwargs = {}
            for n in names:
                if n not in kwargs:
                    raise Exception("Must provide input source for: {}".format(n))
                setattr(self, n, kwargs.get(n))
                self._kwargs[n] = kwargs.get(n)

            self._nodes = [getattr(self, n) for n in names]
            self._lambda = lambdify(syms, expr, modules=modules)(**self._kwargs)
            self._expr = expr

            super(Streaming, self).__init__(output_node=self._lambda)

    return Streaming




class StreamEnd:
    '''Indicates that a stream has nothing left in it'''
    pass


class StreamRepeat:
    '''Indicates that a stream has a gap, this object should be ignored
    and the previous action repeated'''
    pass


class StreamNone:
    '''indicates that a stream does not have a value'''

    def __init__(self, last=None):
        self.value = last

    def all_bin_ops(self, other):
        return self

    def all_un_ops(self):
        return self

    __add__ = all_bin_ops
    __radd__ = all_bin_ops
    __sub__ = all_bin_ops
    __rsub__ = all_bin_ops
    __mul__ = all_bin_ops
    __rmul__ = all_bin_ops
    __div__ = all_bin_ops
    __rdiv__ = all_bin_ops
    __truediv__ = all_bin_ops
    __rtruediv__ = all_bin_ops
    __pow__ = all_bin_ops
    __rpow__ = all_bin_ops
    __mod__ = all_bin_ops
    __rmod__ = all_bin_ops
    __and__ = all_bin_ops
    __or__ = all_bin_ops
    __invert__ = all_un_ops
    def __bool__(self): return False
    def int(self): return 0
    def float(self): return 0
    __lt__ = all_bin_ops
    __le__ = all_bin_ops
    __gt__ = all_bin_ops
    __ge__ = all_bin_ops
    __eq__ = all_bin_ops
    __ne__ = all_bin_ops
    __neg__ = all_un_ops
    __nonzero__ = all_un_ops
    __len__ = all_un_ops



import json as JSON
import websockets
from .output import _OUTPUT_GRAPHVIZSHAPE
from ..base import Node
from ...base import StreamNone, StreamEnd


def WebSocket(node, url='', json=False, wrap=False, field=None, response=False):
    '''Connect to websocket and send data

    Args:
        foo (callable): input stream
        foo_kwargs (dict): kwargs for the input stream
        url (str): websocket url to connect to
        json (bool): dump data as json
        wrap (bool): wrap result in a list
    '''
    async def _send(data, url=url, json=json, wrap=wrap, field=field, response=response):
        if isinstance(data, (StreamNone, StreamEnd)):
            return data

        if wrap:
            data = [data]
        if json:
            data = JSON.dumps(data)

        await ret._websocket.send(data)

        if response:
            msg = await ret._websocket.recv()

        else:
            msg = '{}'

        if json:
            msg = JSON.loads(msg)

        if field:
            msg = msg[field]

        if wrap:
            msg = [msg]

        return msg

    ret = Node(foo=_send, name='WebSocket', inputs=1, graphvizshape=_OUTPUT_GRAPHVIZSHAPE)
    ret._websocket = websockets.connect(url)

    node._downstream.append((ret, 0))
    ret._upstream.append(node)
    return ret
